# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import signal
import socket
import asyncio

from hagworm import hagworm_slogan
from hagworm import __version__ as hagworm_version
from hagworm.extend.asyncio.base import Utils, install_uvloop
from hagworm.extend.interface import RunnableInterface

from ..error import RouterError


class Router:

    def __init__(self, root=r''):

        self._root = root
        self._reg_func = {}

    def __repr__(self):

        return self._reg_func.__repr__()

    def __call__(self, method, *args, **kwargs):

        func = self._reg_func.get(method)

        if not func:
            raise RouterError(f'{method} not exists')

        return func(*args, **kwargs)

    def _reg(self, method, func):

        _method = f'{self._root}{method}'

        if _method in self._reg_func:
            raise RouterError(f'{method} has exists')

        self._reg_func[_method] = func

    def reg(self, method):

        def _reg_func(func):
            self._reg(method, func)
            return func

        return _reg_func

    def items(self):

        return self._reg_func.items()

    def include(self, router):

        for method, func in router.items():
            self._reg(method, func)


class AsyncTcpServer(RunnableInterface):

    def __init__(
            self, client_connected_cb, address, *,
            backlog=None, buffer_limit=0xffffff,
            on_startup=None, on_shutdown=None
    ):

        self._listeners = [(client_connected_cb, address)]

        self._backlog = backlog
        self._buffer_limit = buffer_limit

        self._on_startup = on_startup
        self._on_shutdown = on_shutdown

        self._servers = []
        self._server_tasks = []

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    async def __aenter__(self):

        for server in self._servers:
            await server.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, _traceback):

        for server in self._servers:
            await server.__aexit__(exc_type, exc_value, _traceback)

    def add_listener(self, client_connected_cb, address):

        self._listeners.append((client_connected_cb, address))

    def run(self):

        environment = Utils.environment()

        Utils.log.info(
            f'{hagworm_slogan}'
            f'hagworm {hagworm_version}\n'
            f'python {environment["python"]}\n'
            f'system {" ".join(environment["system"])}'
        )

        install_uvloop()

        asyncio.run(self._run())

    async def _run(self):

        if self._on_startup is not None:
            await self._on_startup()

        for client_connected_cb, address in self._listeners:

            sock = socket.create_server(address, family=socket.AF_INET, backlog=self._backlog, reuse_port=True)

            self._servers.append(
                await asyncio.start_server(client_connected_cb, limit=self._buffer_limit, sock=sock)
            )

            Utils.log.success(f'tcp server [pid:{Utils.getpid()}] startup complete: {sock.getsockname()}')

        async with self:

            for _server in self._servers:
                self._server_tasks.append(
                    asyncio.create_task(_server.serve_forever())
                )

            for _server_task in self._server_tasks:

                try:
                    await _server_task
                except asyncio.CancelledError as _:
                    pass

        if self._on_shutdown is not None:
            await self._on_shutdown()

        Utils.log.success(f'tcp server [pid:{Utils.getpid()}] shutdown')

    def _exit(self, *_):

        for _server_task in self._server_tasks:
            _server_task.cancel()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import re
import struct
import threading

from enum import Enum as _Enum
from io import BytesIO


class Enum(_Enum):

    @classmethod
    def items(cls):
        return cls.to_dict().items()

    @classmethod
    def keys(cls):
        return cls.to_dict().keys()

    @classmethod
    def values(cls):
        return cls.to_dict().values()

    @classmethod
    def to_dict(cls):

        if not hasattr(cls, r'_to_dict'):
            setattr(cls, r'_to_dict', {item.name: item.value for item in cls})

        return getattr(cls, r'_to_dict')

    @classmethod
    def to_keys_dict(cls):

        if not hasattr(cls, r'_to_keys_dict'):
            setattr(cls, r'_to_keys_dict', {item.name: item for item in cls})

        return getattr(cls, r'_to_keys_dict')

    @classmethod
    def to_values_dict(cls):

        if not hasattr(cls, r'_to_values_dict'):
            setattr(cls, r'_to_values_dict', {item.value: item for item in cls})

        return getattr(cls, r'_to_values_dict')

    @classmethod
    def has_key(cls, key):
        return key in cls.to_keys_dict()

    @classmethod
    def has_value(cls, value):
        return value in cls.to_values_dict()


class IntEnum(int, _Enum):
    pass


class StrEnum(str, _Enum):
    pass


class Result(dict):
    """返回结果类
    """

    def __init__(self, code=0, **kwargs):

        super().__init__(code=code, **kwargs)

    def __bool__(self):

        return self.code == 0

    @property
    def code(self):

        return self.get(r'code')


class ThreadList(threading.local):
    """多线程安全的列表
    """

    __slots__ = [r'data']

    def __init__(self):

        self.data = []


class ThreadDict(threading.local):
    """多线程安全的字典
    """

    __slots__ = [r'data']

    def __init__(self):

        self.data = {}


class ByteArrayAbstract:
    """ByteArray抽象类
    """

    NETWORK = r'!'
    NATIVE = r'='
    NATIVE_ALIGNMENT = r'@'
    LITTLE_ENDIAN = r'<'
    BIG_ENDIAN = r'>'

    def __init__(self):

        self._endian = self.NETWORK

    def get_endian(self):

        return self._endian

    def set_endian(self, val):

        self._endian = val

    def read(self, size):

        raise NotImplementedError()

    def write(self, buffer):

        raise NotImplementedError()

    def read_pad_byte(self, _len):

        struct.unpack(f'{self._endian}{_len}x', self.read(_len))

    def write_pad_byte(self, _len):

        self.write(struct.pack(f'{self._endian}{_len}x'))

    def read_char(self):

        return struct.unpack(f'{self._endian}c', self.read(1))[0]

    def write_char(self, val):

        self.write(struct.pack(f'{self._endian}c', val))

    def read_signed_char(self):

        return struct.unpack(f'{self._endian}b', self.read(1))[0]

    def write_signed_char(self, val):

        self.write(struct.pack(f'{self._endian}b', val))

    def read_unsigned_char(self):

        return struct.unpack(f'{self._endian}B', self.read(1))[0]

    def write_unsigned_char(self, val):

        self.write(struct.pack(f'{self._endian}B', val))

    def read_bool(self):

        return struct.unpack(f'{self._endian}?', self.read(1))[0]

    def write_bool(self, val):

        self.write(struct.pack(f'{self._endian}?', val))

    def read_short(self):

        return struct.unpack(f'{self._endian}h', self.read(2))[0]

    def write_short(self, val):

        self.write(struct.pack(f'{self._endian}h', val))

    def read_unsigned_short(self):

        return struct.unpack(f'{self._endian}H', self.read(2))[0]

    def write_unsigned_short(self, val):

        self.write(struct.pack(f'{self._endian}H', val))

    def read_int(self):

        return struct.unpack(f'{self._endian}i', self.read(4))[0]

    def write_int(self, val):

        self.write(struct.pack(f'{self._endian}i', val))

    def read_unsigned_int(self):

        return struct.unpack(f'{self._endian}I', self.read(4))[0]

    def write_unsigned_int(self, val):

        self.write(struct.pack(f'{self._endian}I', val))

    def read_long(self):

        return struct.unpack(f'{self._endian}l', self.read(4))[0]

    def write_long(self, val):

        self.write(struct.pack(f'{self._endian}l', val))

    def read_unsigned_long(self):

        return struct.unpack(f'{self._endian}L', self.read(4))[0]

    def write_unsigned_long(self, val):

        self.write(struct.pack(f'{self._endian}L', val))

    def read_long_long(self):

        return struct.unpack(f'{self._endian}q', self.read(8))[0]

    def write_long_long(self, val):

        self.write(struct.pack(f'{self._endian}q', val))

    def read_unsigned_long_long(self):

        return struct.unpack(f'{self._endian}Q', self.read(8))[0]

    def write_unsigned_long_long(self, val):

        self.write(struct.pack(f'{self._endian}Q', val))

    def read_float(self):

        return struct.unpack(f'{self._endian}f', self.read(4))[0]

    def write_float(self, val):

        self.write(struct.pack(f'{self._endian}f', val))

    def read_double(self):

        return struct.unpack(f'{self._endian}d', self.read(8))[0]

    def write_double(self, val):

        self.write(struct.pack(f'{self._endian}d', val))

    def read_bytes(self, _len):

        return struct.unpack(f'{self._endian}{_len}s', self.read(_len))[0]

    def write_bytes(self, val):

        self.write(struct.pack(f'{self._endian}{len(val)}s', val))

    def read_string(self, _len):

        return self.read_bytes(_len).decode()

    def write_string(self, val):

        self.write_bytes(val.encode())

    def read_pascal_bytes(self, _len):

        return struct.unpack(f'{self._endian}{_len}p', self.read(_len))[0]

    def write_pascal_bytes(self, val):

        self.write(struct.pack(f'{self._endian}{len(val)}p', val))

    def read_pascal_string(self, _len):

        return self.read_pascal_bytes(_len).decode()

    def write_pascal_string(self, val):

        self.write_pascal_bytes(val.encode())

    def read_python_int(self, _len):

        return struct.unpack(f'{self._endian}{_len}P', self.read(_len))[0]

    def write_python_int(self, val):

        self.write(struct.pack(f'{self._endian}{len(val)}P', val))


class ByteArray(BytesIO, ByteArrayAbstract):
    """扩展的BytesIO类
    """

    NETWORK = r'!'
    NATIVE = r'='
    NATIVE_ALIGNMENT = r'@'
    LITTLE_ENDIAN = r'<'
    BIG_ENDIAN = r'>'

    def __init__(self, initial_bytes=None):

        BytesIO.__init__(self, initial_bytes)
        ByteArrayAbstract.__init__(self)


class KeyLowerDict(dict):

    _PATTERN = re.compile(r'(?<=[a-z])([A-Z])')

    def __init__(self, _dict):

        super().__init__(
            {
                KeyLowerDict._PATTERN.sub(r'_\1', key).lower(): KeyLowerDict(val) if isinstance(val, dict) else val
                for key, val in _dict.items()
            }
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import time
import fastapi

from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.exceptions import HTTPException
from starlette.routing import iscoroutinefunction_or_partial
from starlette.concurrency import run_in_threadpool

from fastapi import __version__ as fastapi_version
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.openapi.docs import get_swagger_ui_html

from ... import hagworm_label, hagworm_slogan
from ... import __version__ as hagworm_version

from ...extend.trace import get_trace_id, refresh_trace_id
from ...extend.asyncio.base import Utils, install_uvloop
from ...extend.logging import DEFAULT_LOG_FILE_NAME, DEFAULT_LOG_FILE_ROTATOR, init_logger

from .response import Response, ErrorResponse


get_swagger_ui_html.__kwdefaults__.update(
    {
        r'swagger_js_url': r'https://cdn.bootcdn.net/ajax/libs/swagger-ui/4.14.0/swagger-ui-bundle.min.js',
        r'swagger_css_url': r'https://cdn.bootcdn.net/ajax/libs/swagger-ui/4.14.0/swagger-ui.min.css',
    }
)


DEFAULT_HEADERS = [(r'Server', hagworm_label)]


def create_fastapi(
        log_level=r'info', log_handler=None, log_file_path=None, log_file_name=DEFAULT_LOG_FILE_NAME,
        log_file_rotation=DEFAULT_LOG_FILE_ROTATOR, log_file_retention=0xff, log_extra=None, log_enqueue=False,
        debug=False,
        routes=None,
        **setting
):

    init_logger(
        level=log_level.upper(),
        handler=log_handler,
        file_path=log_file_path,
        file_name=log_file_name,
        file_rotation=log_file_rotation,
        file_retention=log_file_retention,
        extra=log_extra,
        enqueue=log_enqueue,
        debug=debug
    )

    environment = Utils.environment()

    Utils.log.info(
        f'{hagworm_slogan}'
        f'hagworm {hagworm_version}\n'
        f'fastapi {fastapi_version}\n'
        f'python {environment["python"]}\n'
        f'system {" ".join(environment["system"])}'
    )

    setting.setdefault(r'title', r'Hagworm')
    setting.setdefault(r'version', hagworm_version)

    install_uvloop()

    _fastapi = fastapi.FastAPI(debug=debug, routes=routes, **setting)

    _fastapi.add_exception_handler(HTTPException, http_exception_handler)
    _fastapi.add_exception_handler(RequestValidationError, request_validation_exception_handler)

    return _fastapi


class APIRoute(fastapi.routing.APIRoute):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.app = self.request_response()

    async def prepare(self, request):

        pass

    def request_response(self) -> ASGIApp:

        func = self.get_route_handler()

        is_coroutine = iscoroutinefunction_or_partial(func)

        async def app(scope: Scope, receive: Receive, send: Send) -> None:

            request = Request(scope, receive, send, self)

            try:

                trace_id = refresh_trace_id(request.get_header(r'x-request-id', None))

                request_time = time.time() * 1000

                await self.prepare(request)

                if is_coroutine:
                    response = await func(request)
                else:
                    response = await run_in_threadpool(func, request)

                response.headers.update(
                    {
                        r'x-trace-id': trace_id,
                        r'x-server-time-ms': r'{:.0f}'.format(request_time),
                        r'x-request-ttl-ms': r'{:.3f}'.format(time.time() * 1000 - request_time),
                        r'x-request-payload': request.get_header(r'x-request-payload', r''),
                    }
                )

                await response(scope, receive, send)

            except ErrorResponse as err:

                Utils.log.warning(f'ErrorResponse: {request.path}\n{err}\n{request.debug_info}')

                await err(scope, receive, send)

        return app


class APIRouter(fastapi.APIRouter):
    """目录可选末尾的斜杠访问
    """

    def __init__(
            self,
            *,
            prefix=r'',
            default_response_class=Response,
            route_class=APIRoute,
            **kwargs
    ):

        super().__init__(
            prefix=prefix,
            default_response_class=default_response_class,
            route_class=route_class,
            **kwargs
        )

    def _get_path_alias(self, path):

        _path = path.rstrip(r'/')

        if not self.prefix and not _path:
            return [path]

        _path_split = os.path.splitext(_path)

        if _path_split[1]:
            return [_path]

        return [_path, _path + r'/']

    def api_route(self, path, *args, **kwargs):

        def _decorator(func):

            for index, _path in enumerate(self._get_path_alias(path)):

                self.add_api_route(_path, func, *args, **kwargs)

                # 兼容的URL将不会出现在docs中
                if index == 0:
                    kwargs[r'include_in_schema'] = False

            return func

        return _decorator


class Request(fastapi.Request):

    def __init__(self, scope: Scope, receive: Receive, send: Send, api_route: APIRoute):

        super().__init__(scope, receive, send)

        self._api_route = api_route

    @property
    def debug_info(self):

        return {
            'trace_id': get_trace_id(),
            'type': self.scope.get(r'type'),
            'http_version': self.scope.get(r'http_version'),
            'server': self.scope.get(r'server'),
            'client': self.scope.get(r'client'),
            'scheme': self.scope.get(r'scheme'),
            'method': self.scope.get(r'method'),
            'root_path': self.scope.get(r'root_path'),
            'path': self.scope.get(r'path'),
            'query_string': self.scope.get(r'query_string'),
            'headers': self.scope.get(r'headers'),
        }

    @property
    def route(self):

        return self._api_route

    @property
    def path(self):

        return self._api_route.path

    @property
    def tags(self):

        return self._api_route.tags

    @property
    def referer(self):

        return self.headers.get(r'Referer')

    @property
    def client_ip(self):

        if self.x_forwarded_for:
            return self.x_forwarded_for[0]
        else:
            return self.client_host

    @property
    def client_host(self):

        return self.headers.get(r'X-Real-IP', self.client.host)

    @property
    def x_forwarded_for(self):

        return Utils.split_str(self.headers.get(r'X-Forwarded-For', r''), r',')

    @property
    def content_type(self):

        return self.headers.get(r'Content-Type')

    @property
    def content_length(self):

        result = self.headers.get(r'Content-Length', r'')

        return int(result) if result.isdigit() else 0

    def get_header(self, name, default=None):

        return self.headers.get(name, default)


async def http_exception_handler(
        request: Request, exc: HTTPException
) -> ErrorResponse:

    headers = getattr(exc, r'headers', None)

    if headers:
        return ErrorResponse(
            -1,
            content={r'detail': exc.detail},
            status_code=exc.status_code,
            headers=headers
        )
    else:
        return ErrorResponse(
            -1,
            content={r'detail': exc.detail},
            status_code=exc.status_code
        )


async def request_validation_exception_handler(
        request: Request, exc: RequestValidationError
) -> ErrorResponse:

    return ErrorResponse(
        -1,
        content={r'detail': jsonable_encoder(exc.errors())},
        status_code=HTTP_400_BAD_REQUEST,
    )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import asyncio

from tempfile import TemporaryFile

from .base import Utils

from ..interface import ContextManager


class _BufferAbstract:

    def __init__(self, maxsize=0):

        self._buffer: asyncio.Queue = asyncio.Queue(maxsize)
        
        self._consume_task: asyncio.Task = Utils.create_task(
            self._do_consume_task()
        )

    async def _do_consume_task(self):
        raise NotImplementedError()

    def cancel(self):
        self._consume_task.cancel()

    def qsize(self):

        return self._buffer.qsize()

    def append(self, data):

        self._buffer.put_nowait(data)

    async def append_wait(self, data, timeout=0):

        if timeout > 0:
            await asyncio.wait_for(
                self._buffer.put(data),
                timeout
            )
        else:
            await self._buffer.put(data)


class DataQueue(_BufferAbstract):

    def __init__(self, handler, task_limit=1, maxsize=0):

        super().__init__(maxsize)

        self._handler = handler

        self._tasks = set()

        self._task_limit = task_limit
        self._wait_future: asyncio.Future = None

    async def _do_consume_task(self):

        while True:

            try:

                if len(self._tasks) < self._task_limit:

                    data = await self._buffer.get()

                    task = Utils.create_task(
                        self._handler(data)
                    )

                    task.add_done_callback(self._on_task_done)
                    self._tasks.add(task)

                else:

                    self._wait_future = asyncio.Future()
                    await self._wait_future
                    self._wait_future = None

            except Exception as err:

                Utils.log.error(str(err))

    def _on_task_done(self, task):

        if task in self._tasks:
            self._tasks.remove(task)

        if self._wait_future is not None and self._wait_future.done() is False:
            self._wait_future.set_result(None)

    def size(self):

        return len(self._tasks)


class QueueBuffer(DataQueue):

    def __init__(self, handler, slice_size, timeout=1, task_limit=1, maxsize=0):

        super().__init__(handler, task_limit, maxsize)

        self._slice_data = []
        self._slice_size = slice_size
        self._timeout = timeout

    async def _get_slice_data(self):

        for _ in range(self._slice_size):
            self._slice_data.append(
                await self._buffer.get()
            )

    async def _do_consume_task(self):

        while True:

            try:

                if len(self._tasks) < self._task_limit:

                    try:
                        await asyncio.wait_for(
                            self._get_slice_data(),
                            self._timeout
                        )
                    except asyncio.TimeoutError:
                        pass

                    if self._slice_data:

                        task = Utils.create_task(
                            self._handler(self._slice_data.copy())
                        )

                        task.add_done_callback(self._on_task_done)
                        self._tasks.add(task)

                        self._slice_data.clear()

                else:

                    self._wait_future = asyncio.Future()
                    await self._wait_future
                    self._wait_future = None

            except Exception as err:

                Utils.log.error(str(err))

    def qsize(self):

        return super().qsize() + len(self._slice_data)


class FileBuffer(ContextManager):
    """文件缓存类
    """

    def __init__(self, slice_size=0x1000000):

        self._buffers = []

        self._slice_size = slice_size

        self._read_offset = 0

        self._append_buffer()

    def _context_release(self):

        self.close()

    def _append_buffer(self):

        self._buffers.append(TemporaryFile())

    def close(self):

        while len(self._buffers) > 0:
            self._buffers.pop(0).close()

        self._read_offset = 0

    def write(self, data):

        buffer = self._buffers[-1]

        buffer.seek(0, 2)
        buffer.write(data)

        if buffer.tell() >= self._slice_size:
            buffer.flush()
            self._append_buffer()

    def read(self, size=None):

        buffer = self._buffers[0]

        buffer.seek(self._read_offset, 0)

        result = buffer.read(size)

        if len(result) == 0 and len(self._buffers) > 1:
            self._buffers.pop(0).close()
            self._read_offset = 0
        else:
            self._read_offset = buffer.tell()

        return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from consul import Consul

from hagworm.extend.config import ConfigureBase


class Configure(ConfigureBase):
    """配置类
    """

    __slots__ = [r'_client', r'_path']

    def __init__(self, host, port, path=r'config/', **kwargs):

        super().__init__()

        self._client = Consul(host, port, **kwargs)

        self._path = path

    def _get(self, section, key):

        _, data = self._client.kv.get(f'{self._path}{section}/{key}')

        if data and r'Value' in data:
            return data[r'Value'].decode(r'utf-8')
        else:
            return None

    def reload(self):

        for _key, _section, _type in self._key_section.values():

            if _type is str:
                self.__setattr__(_key, self._get(_section, _key))
            elif _type is int:
                self.__setattr__(_key, int(self._get(_section, _key)))
            elif _type is float:
                self.__setattr__(_key, float(self._parser.getfloat(_section, _key)))
            elif _type is bool:
                self.__setattr__(_key, bool(self._parser.getboolean(_section, _key)))
            else:
                self.__setattr__(_key, _type.decode(self._parser.get(_section, _key)))

        return self



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from aliyun.log import QueuedLogHandler
from aliyun.log.logclient import LogClient as _LogClient
from aliyun.log.version import LOGGING_HANDLER_USER_AGENT
from aliyun.log.gethistogramsrequest import GetHistogramsRequest

from ...extend.error import catch_error
from ...extend.asyncio.future import ThreadWorker


_thread_worker = ThreadWorker(32)


class LogClient(_LogClient):

    def __init__(self, endpoint, access_key_id, access_key, project, log_store, **kwargs):

        super().__init__(endpoint, access_key_id, access_key, **kwargs)

        self._project = project
        self._log_store = log_store

    @_thread_worker
    def async_get_histograms(self, from_time, to_time, *, topic=None, query=None):

        result = None

        with catch_error():

            request = GetHistogramsRequest(self._project, self._log_store, from_time, to_time, topic, query)

            respose = self.get_histograms(request)

            if respose.progress == r'Complete':
                result = respose.body

        return result

    @_thread_worker
    def async_get_log(self, from_time, to_time, offset, size, *, topic=None, query=None, reverse=False):

        result = None

        with catch_error():

            respose = self.get_log(
                self._project, self._log_store,
                from_time, to_time, topic, query, reverse, offset, size
            )

            if respose.progress == r'Complete':
                result = respose.body

        return result


class LogHandler(QueuedLogHandler):

    def __init__(self, end_point, access_key_id, access_key, project, log_store, topic=None, source=None, **kwargs):

        super().__init__(end_point, access_key_id, access_key, project, log_store, topic, **kwargs)

        self._source = source

    def create_client(self):

        self.client = LogClient(
            self.end_point, self.access_key_id, self.access_key,
            self.project, self.log_store, source=self._source
        )

        self.client.set_user_agent(LOGGING_HANDLER_USER_AGENT)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import asyncio
import inspect
import aio_pika
import async_timeout

from typing import Dict, Callable

from .publish import RabbitMQProducerForExchangePool
from .consume import RabbitMQConsumerForExchange

from ...extend.error import catch_error
from ...extend.asyncio.base import Utils


class RpcServer:

    def __init__(
            self, url: str, pool_size: int, exchange_name: str, *,
            server_name: str = None, response_timeout: int = 60
    ):

        self._rpc_function: Dict[str, Callable] = {}

        self._producer = RabbitMQProducerForExchangePool(
            url, pool_size, exchange_name,
            exchange_type=aio_pika.ExchangeType.TOPIC
        )

        self._consumer = RabbitMQConsumerForExchange(url)
        self._consumer.config(
            exchange_name, self._consume_message, True,
            queue_config={
                r'name': f'rpc_server.{server_name}' if server_name is not None else r'rpc_server',
                r'arguments': {r'x-message-ttl': int(response_timeout * 1000)}
            }
        )

    async def initialize(self):

        self._rpc_function[r'ping'] = lambda: r'pong'

        await self._producer.connect()
        await self._consumer.connect()

    async def release(self):

        await self._producer.close()
        await self._consumer.close()

        self._rpc_function.clear()

    async def _consume_message(self, message: aio_pika.IncomingMessage):

        with catch_error():

            _message = Utils.msgpack_decode(message.body)

            Utils.log.info(f'rpc client request: {_message}')

            _func_name = _message.get(r'name')
            _func_args = _message.get(r'args')
            _func_kwargs = _message.get(r'kwargs')

            if _func_name in self._rpc_function:

                try:

                    result = self._rpc_function[_func_name](*_func_args, **_func_kwargs)

                    if inspect.isawaitable(result):
                        result = await result

                    response = {
                        r'name': _func_name,
                        r'data': result,
                    }

                except Exception as err:

                    response = {
                        r'name': _func_name,
                        r'error': str(err),
                    }

            else:

                response = {
                    r'name': _func_name,
                    r'error': r'function not found',
                }

            await self._producer.publish(
                aio_pika.Message(
                    body=Utils.msgpack_encode(response),
                    correlation_id=message.correlation_id
                ),
                routing_key=message.reply_to
            )

    def register(self, name: str, func: Callable):

        self._rpc_function[name] = func

        Utils.log.info(f'rpc server register {name} {func}')


class RpcClient:

    def __init__(
            self, url: str, pool_size: int, exchange_name: str, *,
            request_timeout: int = 60
    ):

        self._futures: Dict[str, asyncio.Future] = {}
        self._request_timeout = request_timeout

        self._producer = RabbitMQProducerForExchangePool(
            url, pool_size, exchange_name,
            exchange_type=aio_pika.ExchangeType.TOPIC
        )

        self._consumer = RabbitMQConsumerForExchange(url)
        self._consumer.config(
            exchange_name, self._consume_message, True,
            queue_config={
                r'name': f'rpc_client.{Utils.uuid1()}',
                r'exclusive': True,
                r'arguments': {r'x-message-ttl': int(request_timeout * 1000)}
            }
        )

    async def initialize(self):

        await self._producer.connect()
        await self._consumer.connect()

    async def release(self):

        await self._producer.close()
        await self._consumer.close()

    def _consume_message(self, message: aio_pika.IncomingMessage):

        with catch_error():

            response = Utils.msgpack_decode(message.body)

            if message.correlation_id in self._futures:

                self._futures.get(
                    message.correlation_id
                ).set_result(
                    response
                )

    async def call(self, name: str, *, args=None, kwargs=None, server_name: str = None):

        correlation_id = Utils.uuid1()

        future = self._futures[correlation_id] = asyncio.Future()

        message = {
            r'name': name,
            r'args': args if args is not None else [],
            r'kwargs': kwargs if kwargs is not None else {},
        }

        try:

            async with async_timeout.timeout(self._request_timeout):

                await self._producer.publish(
                    aio_pika.Message(
                        Utils.msgpack_encode(message),
                        correlation_id=correlation_id,
                        reply_to=self._consumer.queue_name,
                    ),
                    routing_key=f'rpc_server.{server_name}' if server_name is not None else r'rpc_server',
                )

                response = await future

        except Exception as err:

            Utils.log.error(f'rpc call timeout: {message}')

            raise err

        finally:

            del self._futures[correlation_id]

        if r'error' in response:
            raise Exception(response[r'error'])

        return response[r'data']



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from mq_http_sdk.mq_client import MQClient as _MQClient
from mq_http_sdk.mq_consumer import MQConsumer as _MQConsumer
from mq_http_sdk.mq_producer import MQProducer as _MQProducer, MQTransProducer as _MQTransProducer, TopicMessage
from mq_http_sdk.mq_exception import MQServerException, MQClientNetworkException

from ...extend.interface import TaskInterface
from ...extend.asyncio.base import Utils, AsyncCirculator
from ...extend.asyncio.task import IntervalTask
from ...extend.asyncio.future import ThreadPool


CLIENT_WAIT_TIMEOUT = 10
CLIENT_ERROR_RETRY_COUNT = 5


class MQConsumer(_MQConsumer):

    async def consume_message(self, batch_size=1, wait_seconds=-1):

        return await self.mq_client.thread_run(super().consume_message, batch_size, wait_seconds)

    async def ack_message(self, messages):

        receipt_handle_list = [msg.receipt_handle for msg in messages]

        return await self.mq_client.thread_run(super().ack_message, receipt_handle_list)


class MQProducer(_MQProducer):

    async def publish_message(self, msg_body, msg_tag=r''):

        return await self.mq_client.thread_run(super().publish_message, TopicMessage(msg_body, msg_tag))


class MQTransProducer(_MQTransProducer):

    async def publish_message(self, msg_body, msg_tag=r''):

        return await self.mq_client.thread_run(super().publish_message, TopicMessage(msg_body, msg_tag))

    async def consume_half_message(self, batch_size=1, wait_seconds=-1):

        return await self.mq_client.thread_run(super().consume_half_message, batch_size, wait_seconds)

    async def commit(self, message):

        return await self.mq_client.thread_run(super().commit, message.receipt_handle)

    async def rollback(self, message):

        return await self.mq_client.thread_run(super().rollback, message.receipt_handle)


class _TaskMixin(TaskInterface):

    def __init__(self, message_handler):

        self._task = IntervalTask.create(1, self._do_task)

        self._message_handler = message_handler

    def start(self):
        self._task.start()

    def stop(self):
        self._task.stop()

    def is_running(self):
        self._task.is_running()

    async def _do_task(self):
        raise NotImplementedError()


class MQClient(_MQClient):

    def __init__(self, host, access_id, access_key, security_token=r'', debug=False, logger=None):

        super().__init__(host, access_id, access_key, security_token, debug, logger)

        # 原版SDK中的MQClient线程不安全，利用线程池仅有一个线程这种特例，保证线程安全
        self._thread_pool = ThreadPool(1)

    async def thread_run(self, func, *args, **kwargs):

        global CLIENT_ERROR_RETRY_COUNT

        result = None

        # 增加网络异常重试机制
        async for times in AsyncCirculator(max_times=CLIENT_ERROR_RETRY_COUNT):

            try:

                result = await self._thread_pool.run(func, *args, **kwargs)

            except MQClientNetworkException as err:

                if times < CLIENT_ERROR_RETRY_COUNT:
                    Utils.log.warning(err)
                else:
                    raise err

            else:

                break

        return result

    def get_consumer(self, instance_id, topic_name, consumer, message_tag=r''):

        return MQConsumer(instance_id, topic_name, consumer, message_tag, self, self.debug)

    def get_producer(self, instance_id, topic_name):

        return MQProducer(instance_id, topic_name, self, self.debug)

    def get_trans_producer(self, instance_id, topic_name, group_id):

        return MQTransProducer(instance_id, topic_name, group_id, self, self.debug)


class MQMultiProducers:

    def __init__(self, client_num, host, access_id, access_key, security_token=r'', debug=False, logger=None):

        self._clients = [
            MQClient(host, access_id, access_key, security_token, debug, logger)
            for _ in range(client_num)
        ]

        self._producers = None
        self._producers_iter = None

    def init_producers(self, instance_id, topic_name):

        self._producers = [
            client.get_producer(instance_id, topic_name)
            for client in self._clients
        ]

        self._producers_iter = Utils.itertools.cycle(self._producers)

    async def publish_message(self, msg_body, msg_tag=r''):

        producer = next(self._producers_iter)

        return await producer.publish_message(msg_body, msg_tag)


class MQCycleConsumer(MQConsumer, _TaskMixin):

    def __init__(self, message_handler, instance_id, topic_name, consumer, message_tag, mq_client, debug=False):

        MQConsumer.__init__(self, instance_id, topic_name, consumer, message_tag, mq_client, debug)
        _TaskMixin.__init__(self, message_handler)

        self._client_batch_size = 1
        self._client_wait_seconds = CLIENT_WAIT_TIMEOUT

    def set_client_batch_size(self, val):

        self._client_batch_size = val

    def set_client_wait_seconds(self, val):

        self._client_wait_seconds = val

    async def _do_task(self):

        messages = None

        try:
            messages = await self.consume_message(self._client_batch_size, self._client_wait_seconds)
        except MQServerException as err:
            if err.type != r'MessageNotExist':
                Utils.log.error(err)

        if messages:
            await Utils.awaitable_wrapper(
                self._message_handler(self, messages)
            )


class MQMultiCycleConsumers(TaskInterface):

    def __init__(self, client_num, host, access_id, access_key, security_token=r'', debug=False, logger=None):

        self._clients = [
            MQClient(host, access_id, access_key, security_token, debug, logger)
            for _ in range(client_num)
        ]

        self._consumers = None

        self._running = False

    def init_consumers(self,
                       message_handler, instance_id, topic_name, consumer, message_tag, debug=False,
                       *, consumer_cls=MQCycleConsumer
                       ):

        if self._consumers:
            self.stop()

        self._consumers = [
            consumer_cls(message_handler, instance_id, topic_name, consumer, message_tag, client, debug)
            for client in self._clients
        ]

    def set_client_batch_size(self, val):

        if self._consumers:
            for consumer in self._consumers:
                consumer.set_client_batch_size(val)

    def set_client_wait_seconds(self, val):

        if self._consumers:
            for consumer in self._consumers:
                consumer.set_client_wait_seconds(val)

    def start(self):

        if self._running or self._consumers is None:
            return False

        self._running = True

        for consumer in self._consumers:
            consumer.start()

        return True

    def stop(self):

        if not self._running or self._consumers is None:
            return False

        self._running = False

        for consumer in self._consumers:
            consumer.stop()

        return True

    def is_running(self):

        return self._running


class MQCycleTransProducer(MQTransProducer, _TaskMixin):

    def __init__(self, message_handler, instance_id, topic_name, group_id, mq_client, debug=False):

        MQTransProducer.__init__(self, instance_id, topic_name, group_id, mq_client, debug)
        _TaskMixin.__init__(self, message_handler)

        self._client_batch_size = 1
        self._client_wait_seconds = CLIENT_WAIT_TIMEOUT

    def set_client_batch_size(self, val):

        self._client_batch_size = val

    def set_client_wait_seconds(self, val):

        self._client_wait_seconds = val

    async def _do_task(self):

        messages = None

        try:
            messages = await self.consume_half_message(self._client_batch_size, self._client_wait_seconds)
        except MQServerException as err:
            if err.type != r'MessageNotExist':
                Utils.log.error(err)

        if messages:
            await Utils.awaitable_wrapper(
                self._message_handler(self, messages)
            )


class MQMultiCycleTransProducers(TaskInterface):

    def __init__(self, client_num, host, access_id, access_key, security_token=r'', debug=False, logger=None):

        self._clients = [
            MQClient(host, access_id, access_key, security_token, debug, logger)
            for _ in range(client_num)
        ]

        self._producers = None
        self._producers_iter = None

        self._running = False

    def init_producers(self,
                       message_handler, instance_id, topic_name, group_id, debug=False,
                       *, producer_cls=MQCycleTransProducer
                       ):

        if self._producers:
            self.stop()

        self._producers = [
            producer_cls(message_handler, instance_id, topic_name, group_id, client, debug)
            for client in self._clients
        ]

        self._producers_iter = Utils.itertools.cycle(self._producers)

    def set_client_batch_size(self, val):

        if self._producers:
            for producer in self._producers:
                producer.set_client_batch_size(val)

    def set_client_wait_seconds(self, val):

        if self._producers:
            for producer in self._producers:
                producer.set_client_wait_seconds(val)

    async def publish_message(self, msg_body, msg_tag=r''):

        producer = next(self._producers_iter)

        return await producer.publish_message(msg_body, msg_tag)

    def start(self):

        if self._running or self._producers is None:
            return False

        self._running = True

        for producer in self._producers:
            producer.start()

        return True

    def stop(self):

        if not self._running or self._producers is None:
            return False

        self._running = False

        for producer in self._producers:
            producer.stop()

        return True

    def is_running(self):

        return self._running



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys

import argparse
import py_compile

from .base import Utils


def compile_pyc(file_path, pyc_file_path=None, exclude=None):
    """PYC编译函数
    """

    if exclude:
        exclude = [os.path.join(file_path, val) for val in Utils.split_str(exclude, r'|')]
    else:
        exclude = []

    for root, _, files in os.walk(file_path):

        for _path in exclude:

            if root.find(_path) == 0:
                break

        else:

            for _file in files:

                _, ext_name = os.path.splitext(_file)

                if ext_name != r'.py':
                    continue

                if pyc_file_path is None:

                    dest_path = file_path

                else:

                    dest_path = root.replace(file_path, pyc_file_path)

                    if not os.path.exists(dest_path):
                        os.makedirs(dest_path)

                ori_path = os.path.join(root, _file)
                dest_path = os.path.join(dest_path, _file) + r'c'

                py_compile.compile(ori_path, dest_path, optimize=2)

                sys.stdout.write(f'{ori_path} => {dest_path}\n')


def main():

    result = 0

    parser = argparse.ArgumentParser()

    parser.add_argument(r'-i', r'--input', default=r'./', dest=r'input')
    parser.add_argument(r'-o', r'--output', default=None, dest=r'output')
    parser.add_argument(r'-e', r'--exclude', default=None, dest=r'exclude')

    args = parser.parse_args()

    try:
        compile_pyc(args.input, args.output, args.exclude)
    except Exception as error:
        sys.stderr.write(f'{error}\n')

    return result


if __name__ == r'__main__':

    sys.exit(main())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from .base import Utils

from ..interface import AsyncContextManager
from ..transaction import TransactionAbstract


class Transaction(TransactionAbstract, AsyncContextManager):
    """事务对象

    使用异步上下文实现的一个事务对象，可以设置commit和rollback回调
    未显示commit的情况下，会自动rollback

    """

    async def _context_release(self):

        await self.rollback()

    async def commit(self):

        if self._commit_callbacks is None:
            return

        callbacks = self._commit_callbacks.copy()

        self._clear_callbacks()

        for _callable in callbacks:
            try:
                await Utils.awaitable_wrapper(
                    _callable()
                )
            except Exception as err:
                Utils.log.critical(f'transaction commit error:\n{err}')

    async def rollback(self):

        if self._rollback_callbacks is None:
            return

        callbacks = self._rollback_callbacks.copy()

        self._clear_callbacks()

        for _callable in callbacks:
            try:
                await Utils.awaitable_wrapper(
                    _callable()
                )
            except Exception as err:
                Utils.log.critical(f'transaction rollback error:\n{err}')



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import aio_pika


class RabbitMQConsumer(aio_pika.RobustConnection):
    """RabbitMQ消费者
    """

    def __init__(self, url: str, **kwargs):

        super().__init__(url, **kwargs)

        self._channel: aio_pika.abc.AbstractRobustChannel = None
        self._queue: aio_pika.abc.AbstractRobustQueue = None

        self._channel_qos_config = None

        self._queue_name = None
        self._queue_config = None

        self._consume_func = None
        self._consume_no_ack = None

    @property
    def queue_name(self):

        return self._queue_name

    @property
    def current_channel(self) -> aio_pika.abc.AbstractRobustChannel:

        return self._channel

    @property
    def current_queue(self) -> aio_pika.abc.AbstractRobustQueue:

        return self._queue

    def config(
            self, queue_name, consume_func, consume_no_ack=False,
            *, channel_qos_config=None, queue_config=None
    ):

        self._channel_qos_config = channel_qos_config if channel_qos_config else {r'prefetch_count': 1}

        self._queue_name = queue_name
        self._queue_config = queue_config if queue_config else {}

        self._consume_func = consume_func
        self._consume_no_ack = consume_no_ack

    async def connect(self, timeout: aio_pika.abc.TimeoutType = None) -> None:

        self._RobustConnection__channels.clear()

        await super().connect(timeout)
        await self.ready()

        self._channel = await self.channel()

        await self._channel.set_qos(**self._channel_qos_config)

        self._queue = await self._channel.declare_queue(self._queue_name, **self._queue_config)

        if self._consume_func is not None:
            await self._queue.consume(self._consume_func, no_ack=self._consume_no_ack)

    async def close(self):

        await self._channel.close()
        await super().close()

    async def get(self, *, no_ack=False, timeout=1):

        await self._queue.get(no_ack=no_ack, timeout=timeout)


class RabbitMQConsumerForExchange(RabbitMQConsumer):
    """RabbitMQ注册到交换机的消费者(默认生成排它队列注册)
    """

    def __init__(self, url: str, **kwargs):
        
        super().__init__(url, **kwargs)

        self._exchange: aio_pika.abc.AbstractExchange = None

        self._exchange_name = None
        self._routing_key = None

    @property
    def current_exchange(self) -> aio_pika.abc.AbstractExchange:

        return self._exchange

    def config(
            self, exchange_name, consume_func, consume_no_ack=False,
            *, channel_qos_config=None, queue_config=None, routing_key=None
    ):

        queue_name = None

        if queue_config is None:
            queue_config = {r'exclusive': True}
        elif r'name' in queue_config:
            queue_name = queue_config.pop(r'name')

        super().config(
            queue_name, consume_func, consume_no_ack,
            channel_qos_config=channel_qos_config, queue_config=queue_config
        )

        self._exchange_name = exchange_name
        self._routing_key = routing_key

    async def connect(self, timeout: aio_pika.abc.TimeoutType = None) -> None:

        self._exchange = None

        await super().connect(timeout)

        self._exchange = await self._channel.get_exchange(self._exchange_name)

        await self._queue.bind(self._exchange, self._routing_key)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from cachetools import TTLCache

from .base import Utils
from .transaction import Transaction


class StackCache:
    """堆栈缓存

    使用运行内存作为高速缓存，可有效提高并发的处理能力

    """

    def __init__(self, maxsize=0xff, ttl=60):

        self._cache = TTLCache(maxsize, ttl)

    def has(self, key):

        return key in self._cache

    def get(self, key, default=None):

        return self._cache.get(key, default)

    def set(self, key, val):

        self._cache[key] = val

    def incr(self, key, val=1):

        res = self.get(key, 0) + val

        self.set(key, res)

        return res

    def decr(self, key, val=1):

        res = self.get(key, 0) - val

        self.set(key, res)

        return res

    def delete(self, key):

        del self._cache[key]

    def size(self):

        return len(self._cache)

    def clear(self):

        return self._cache.clear()


class PeriodCounter:

    MIN_EXPIRE = 60

    def __init__(self, time_slice, key_prefix=r'', maxsize=0xffff):

        self._time_slice = time_slice
        self._key_prefix = key_prefix

        # 缓存对象初始化，key最小过期时间60秒
        self._cache = StackCache(maxsize, max(time_slice, self.MIN_EXPIRE))

    def _get_key(self, key=None):

        time_period = Utils.math.floor(Utils.timestamp() / self._time_slice)

        if key is None:
            return f'{self._key_prefix}_{time_period}'
        else:
            return f'{self._key_prefix}_{key}_{time_period}'

    def get(self, key=None):

        _key = self._get_key(key)

        return self._cache.get(_key, 0)

    def incr(self, val, key=None):

        _key = self._get_key(key)

        return self._cache.incr(_key, val)

    def incr_with_trx(self, val, key=None):

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.decr, _key, val)

        return self._cache.incr(_key, val), trx

    def decr(self, val, key=None):

        _key = self._get_key(key)

        return self._cache.decr(_key, val)

    def decr_with_trx(self, val, key=None):

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.incr, _key, val)

        return self._cache.decr(_key, val), trx



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import threading
import traceback


class SafeSingletonMetaclass(type):
    """线程安全的单例的元类实现
    """

    def __init__(cls, _what, _bases=None, _dict=None):

        super().__init__(_what, _bases, _dict)

        cls._instance = None

        cls._lock = threading.Lock()

    def __call__(cls, *args, **kwargs):

        result = None

        cls._lock.acquire()

        try:

            if cls._instance is not None:
                result = cls._instance
            else:
                result = cls._instance = super().__call__(*args, **kwargs)

        except Exception as _:

            traceback.print_exc()

        finally:

            cls._lock.release()

        return result


class SafeSingleton(metaclass=SafeSingletonMetaclass):
    """线程安全的单例基类
    """
    pass


class SingletonMetaclass(type):
    """单例的元类实现
    """

    def __init__(cls, _what, _bases=None, _dict=None):

        super().__init__(_what, _bases, _dict)

        cls._instance = None

    def __call__(cls, *args, **kwargs):

        result = None

        try:

            if cls._instance is not None:
                result = cls._instance
            else:
                result = cls._instance = super().__call__(*args, **kwargs)

        except Exception as _:

            traceback.print_exc()

        return result


class Singleton(metaclass=SingletonMetaclass):
    """单例基类
    """
    pass



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import asyncio
import aio_pika

from typing import List

from ...extend.base import Utils
from ...extend.asyncio.pool import ObjectPool


class RabbitMQProducer(aio_pika.RobustConnection):
    """RabbitMQ发布者
    """

    def __init__(self, url, **kwargs):

        super().__init__(url, **kwargs)

        self._channel: aio_pika.abc.AbstractRobustChannel = None

        self._lock: asyncio.Lock = asyncio.Lock()

    @property
    def current_channel(self) -> aio_pika.abc.AbstractRobustChannel:

        return self._channel

    async def connect(
            self, *,
            channel_number: int = None, publisher_confirms: bool = True, on_return_raises: bool = False,
            timeout: aio_pika.abc.TimeoutType = None
    ):

        await super().connect(timeout)

        await self.ready()

        if self._channel is None:
            self._channel = await self.channel(channel_number, publisher_confirms, on_return_raises)

    async def close(self):

        await self._channel.close()
        await super().close()

    async def publish(self, message, routing_key, **kwargs):

        async with self._lock:
            return await self._channel.default_exchange.publish(
                message if isinstance(message, aio_pika.Message) else aio_pika.Message(message),
                routing_key, **kwargs
            )


class RabbitMQProducerForExchange(RabbitMQProducer):
    """RabbitMQ交换机发布者
    """

    def __init__(self, url, **kwargs):

        super().__init__(url, **kwargs)

        self._exchange: aio_pika.abc.AbstractExchange = None

        self._exchange_name = None
        self._exchange_type = None
        self._exchange_config = None

    @property
    def current_exchange(self) -> aio_pika.abc.AbstractExchange:

        return self._exchange

    def config(
            self, exchange_name, exchange_type=aio_pika.ExchangeType.FANOUT,
            *, exchange_config=None
    ):

        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._exchange_config = exchange_config if exchange_config else {}

    async def connect(
            self, *,
            channel_number: int = None, publisher_confirms: bool = True, on_return_raises: bool = False,
            timeout: aio_pika.abc.TimeoutType = None
    ):

        await super().connect(
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises,
            timeout=timeout
        )

        if self._exchange is None:
            self._exchange = await self._channel.declare_exchange(
                self._exchange_name, self._exchange_type, **self._exchange_config
            )

    async def publish(self, message, routing_key=r'', **kwargs):

        async with self._lock:
            return await self._exchange.publish(
                message if isinstance(message, aio_pika.Message) else aio_pika.Message(message),
                routing_key, **kwargs
            )


class RabbitMQProducerPool(ObjectPool):
    """RabbitMQ发布者连接池
    """

    def __init__(self, url, pool_size, *, connection_config=None):

        self._mq_url = url

        self._connection_config = connection_config if connection_config else {}

        self._connections: List[RabbitMQProducer] = []

        super().__init__(pool_size)

    def _create_obj(self):

        connection = RabbitMQProducer(self._mq_url, **self._connection_config)

        self._connections.append(connection)

        return connection

    async def connect(
            self, *,
            publisher_confirms: bool = True, on_return_raises: bool = False,
            timeout: aio_pika.abc.TimeoutType = None
    ):

        for connection in self._connections:
            await connection.connect(
                publisher_confirms=publisher_confirms,
                on_return_raises=on_return_raises,
                timeout=timeout,
            )

        Utils.log.info(f'rabbitmq producer pool connected: {self._queue.qsize()}')

    async def close(self):

        for connection in self._connections:
            await connection.close()

    async def publish(self, message, routing_key=r'', **kwargs):

        async with self.get() as connection:
            return await connection.publish(
                message if isinstance(message, aio_pika.Message) else aio_pika.Message(message),
                routing_key, **kwargs
            )


class RabbitMQProducerForExchangePool(RabbitMQProducerPool):
    """RabbitMQ交换机发布者连接池
    """

    def __init__(
            self, url, pool_size, exchange_name,
            *, exchange_type=aio_pika.ExchangeType.FANOUT, exchange_config=None, connection_config=None
    ):

        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._exchange_config = exchange_config

        super().__init__(url, pool_size, connection_config=connection_config)

    def _create_obj(self):

        connection = RabbitMQProducerForExchange(self._mq_url, **self._connection_config)
        connection.config(self._exchange_name, self._exchange_type, exchange_config=self._exchange_config)

        self._connections.append(connection)

        return connection



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import traceback

from .base import Utils
from .error import Ignore


class FunctorInterface:
    """仿函数接口定义
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class RunnableInterface:
    """Runnable接口定义
    """

    def run(self, *args, **kwargs):
        raise NotImplementedError()


class TaskInterface:
    """Task接口定义
    """

    def start(self, *args, **kwargs):
        raise NotImplementedError()

    def stop(self, *args, **kwargs):
        raise NotImplementedError()

    def is_running(self, *args, **kwargs):
        raise NotImplementedError()


class ObjectFactoryInterface:
    """对象工厂类接口定义
    """

    def create(self, *args, **kwargs):
        raise NotImplementedError()


class ContextManager:
    """上下文资源管理器

    子类通过实现_context_release接口，方便的实现with语句管理上下文资源释放

    """

    def __enter__(self):

        self._context_initialize()

        return self

    def __exit__(self, exc_type, exc_value, _traceback):

        self._context_release()

        if exc_type and issubclass(exc_type, Ignore):

            return not exc_value.throw()

        elif exc_value:

            Utils.log.error(traceback.format_exc())

            return True

    def _context_initialize(self):

        pass

    def _context_release(self):

        raise NotImplementedError()


class AsyncContextManager:
    """异步上下文资源管理器

    子类通过实现_context_release接口，方便的实现with语句管理上下文资源释放

    """

    async def __aenter__(self):

        await self._context_initialize()

        return self

    async def __aexit__(self, exc_type, exc_value, _traceback):

        await self._context_release()

        if exc_type and issubclass(exc_type, Ignore):

            return not exc_value.throw()

        elif exc_value:

            Utils.log.error(traceback.format_exc())

            return True

    async def _context_initialize(self):

        pass

    async def _context_release(self):

        raise NotImplementedError()
